#!/usr/bin/env python
# coding: utf-8
from __future__ import absolute_import, print_function, unicode_literals

import sys
import os
import argparse
import collections
import datetime
from pprint import pprint

from MySQLdb import ProgrammingError

from hammers import MySqlArgs, query

RESOURCE_QUERY = {
    'ip': query.owned_ip_single,
    'port': query.owned_ports_single,
}

RESOURCE_DELETE_COMMAND = {
    'ip': 'floatingip-delete',
    'port': 'port-delete',
}

assert set(RESOURCE_QUERY) == set(RESOURCE_DELETE_COMMAND)


def normalize_project_name(s):
    return s.lower().replace('-', '').strip()


def days_past(dt):
    return (datetime.datetime.utcnow() - dt).total_seconds() / (60*60*24)


def reaper(db, type_, idle_days, whitelist, describe=False):
    try:
        future_projects = {
            normalize_project_name(row['project_id'])
            for row
            in query.future_reservations(db)
        }
    except ProgrammingError as e:
        if "Table 'blazar.leases' doesn't exist" in str(e):
            # assume we're on KVM, we can't check this.
            future_projects = set()
        else:
            raise

    too_idle_projects = [
        proj
        for proj
        in query.idle_projects(db)
        if (
            days_past(proj['latest_deletion']) > idle_days
            and normalize_project_name(proj['id']) not in future_projects
            and normalize_project_name(proj['id']) not in whitelist
            and normalize_project_name(proj['name']) not in whitelist
        )
    ]
    too_idle_project_ids = [proj['id'] for proj in too_idle_projects]

    # for ip in owned_ips(db, too_idle_project_ids):
        # print('floatingip-delete {}'.format(ip['id']))
    resource_query = RESOURCE_QUERY[type_]
    command = RESOURCE_DELETE_COMMAND[type_]

    if not describe:
        for proj_id in too_idle_project_ids:
            for resource in resource_query(db, proj_id):
                print('{} {}'.format(command, resource['id']))
    else:
        projects = collections.defaultdict(dict)
        for proj_id in too_idle_project_ids:
            for resource in resource_query(db, proj_id):
                assert proj_id == resource.pop('project_id')
                projects[proj_id][resource.pop('id')] = resource
        print('Format: {project_id: {resource_id: {INFO} ...}, ...}\n')
        pprint(dict(projects))
        # print(json.dumps(dict(projects), indent=4))


def main(argv=None):
    if argv is None:
        argv = sys.argv

    parser = argparse.ArgumentParser(description='Floating IP reclaimer. Run '
        'by itself as a "dry run," or pipe it to the Neutron client.')
    mysqlargs = MySqlArgs({
        'user': 'root',
        'password': '',
        'host': 'localhost',
        'port': 3306,
    })
    mysqlargs.inject(parser)

    parser.add_argument('-w', '--whitelist', type=str,
        help='File of project/tenant IDs/names to ignore, one per line. '
             'Ignores case and dashes.')
    parser.add_argument('-i', '--info', action='store_true',
        help='Rather than print Neutron commands, print out info about them.')
    parser.add_argument('type', choices=list(RESOURCE_QUERY),
        help='Grab floating IPs or ports?')
    parser.add_argument('idle_days', type=float,
        help='Number of days since last active instance in project was '
        'deleted to consider it idle.')

    args = parser.parse_args(argv[1:])
    mysqlargs.extract(args)

    whitelist = set()
    if args.whitelist:
        with open(args.whitelist) as f:
            whitelist = {normalize_project_name(line) for line in f}

    db = mysqlargs.connect()

    reaper(db, args.type, args.idle_days, whitelist, args.info)


if __name__ == '__main__':
    sys.exit(main(sys.argv))



# coding: utf-8
from __future__ import absolute_import, print_function, unicode_literals

import itertools

import MySQLdb

__all__ = ['MySqlShim']


class MySqlShim(object):
    batch_size = 100
    limit = 1000

    def __init__(self, **connect_args):
        self.db = MySQLdb.connect(**connect_args)
        self.cursor = self.db.cursor()

    def columns(self):
        return [cd[0] for cd in self.cursor.description]

    def query(self, *cargs, **ckwargs):
        limit = ckwargs.pop('limit', self.limit)

        if ckwargs.pop('immediate', False):
            return list(itertools.islice(self._query(*cargs, **ckwargs), limit))
        else:
            return itertools.islice(self._query(*cargs, **ckwargs), limit)

    def _query(self, *cargs, **ckwargs):
        self.cursor.execute(*cargs, **ckwargs)
        fields = self.columns()
        rows = self.cursor.fetchmany(self.batch_size)
        while rows:
            for row in rows:
                yield dict(zip(fields, row))
            rows = self.cursor.fetchmany(self.batch_size)



# coding: utf-8
from __future__ import absolute_import, print_function, unicode_literals
import functools

QUERIES = {}
def query(q):
    global QUERIES
    QUERIES[q.__name__] = {'f': q}
    return q


@query
def idle_projects(db):
    '''
    Returns rows enumerating all projects that are currently idle (number
    of running instances = 0). Also provides since when the project has been
    idle (when the latest running instance was deleted)
    '''
    sql = '''
    SELECT project.id
         , project.name
        #  , Count(ip.tenant_id)
         , (SELECT deleted_at
            FROM   nova.instances AS instance
            WHERE  instance.project_id = project.id
            ORDER  BY deleted_at DESC
            LIMIT  1)
                AS latest_deletion
    FROM   neutron.floatingips AS ip
       ,   keystone.project AS project
    WHERE  ip.tenant_id = project.id
           AND ip.status = "down"
           AND (SELECT Count(*)
                FROM   nova.instances
                    AS instance
                WHERE  instance.project_id = project.id
                       AND deleted_at IS NULL
                       AND vm_state != "deleted"
                       AND vm_state != "error") = 0
    GROUP  BY tenant_id
    ORDER  BY Count(tenant_id) DESC;
    '''
    return db.query(sql, limit=None)


@query
def owned_ips(db, project_ids):
    '''
    Return all IPs associated with *project_ids*

    Maria 5.5 in production doesn't seem to like this, but works fine with
    a local MySQL 5.7. Is it Maria? 5.5? Too many? See owned_ip_single for one
    that works, but need to call multiple times.
    '''
    sql = '''
    SELECT id
         , status
         , tenant_id AS project_id
    FROM   neutron.floatingips
    WHERE  tenant_id IN %s;
    '''
    return db.query(sql, args=[project_ids], limit=None)


@query
def owned_ip_single(db, project_id):
    '''
    Return all IPs associated with *project_id*
    '''
    sql = '''
    SELECT id
         , status
         , tenant_id AS project_id
    FROM   neutron.floatingips
    WHERE  tenant_id = %s;
    '''
    return db.query(sql, args=[project_id], limit=None)


@query
def projects_with_unowned_ports(db):
    sql = '''
    SELECT tenant_id AS project_id
         , count(tenant_id) AS count_blank_owner
    FROM   neutron.ports
    WHERE  device_owner = ''
    GROUP  BY tenant_id;
    '''
    return db.query(sql, limit=None)


@query
def owned_ports_single(db, project_id):
    sql = '''
    SELECT id
         , status
         , tenant_id AS project_id
    FROM   neutron.ports
    WHERE  tenant_id = %s;
    '''
    return db.query(sql, args=[project_id], limit=None)


@query
def future_reservations(db):
    '''
    Get project IDs with lease end dates in the future that haven't
    been deleted. This will also grab *active* leases, but that's erring
    on the safe side.
    '''
    sql = '''
    SELECT DISTINCT project_id
    FROM   blazar.leases
    WHERE  end_date > Now()
           AND deleted_at is NULL;
    '''
    return db.query(sql, limit=None)


def main(argv):
    import sys
    import argparse
    import ast

    from .mysqlargs import MySqlArgs

    parser = argparse.ArgumentParser(description='Run queries!')
    mysqlargs = MySqlArgs({
        'user': 'root',
        'password': '',
        'host': 'localhost',
        'port': 3306,
    })
    mysqlargs.inject(parser)

    parser.add_argument('query', type=str, choices=QUERIES,
        help='Query to run.',
    )
    parser.add_argument('qargs', type=str, nargs='*',
        help='Arguments for the query (if needed)'
    )

    args = parser.parse_args(argv[1:])
    mysqlargs.extract(args)

    db = mysqlargs.connect()

    # qargs = [ast.literal_eval(a) for a in args.qargs]

    try:
        for row in QUERIES[args.query]['f'](db, *args.qargs):
            print(row)
    except TypeError as e:
        if '{}() takes'.format(args.query) in str(e):
            print('Invalid number of arguments provided to query: {}'.format(str(e)), file=sys.stderr)
            return -1
        else:
            raise


if __name__ == '__main__':
    import sys
    sys.exit(main(sys.argv))



# coding: utf-8
from __future__ import absolute_import, print_function, unicode_literals

import os
import codecs
import glob
import logging
import stat
CP_MODERN = True
try:
    import configparser
except ImportError:
    try:
        from backports import configparser
    except ImportError:
        import io
        import ConfigParser as configparser
        CP_MODERN = False

__all__ = ['MyCnf', 'MYCNF_PATHS']

KEYERROR_LIKE_OPTIONERRORS = (
    configparser.NoSectionError,
    configparser.NoOptionError,
)

# https://dev.mysql.com/doc/refman/5.7/en/option-files.html
# https://mariadb.com/kb/en/mariadb/configuring-mariadb-with-mycnf/
MYCNF_PATHS = [
    '/etc/my.cnf',
    '/etc/mysql/my.cnf',
#     'SYSCONFDIR/my.cnf',
#     'defaults-extra-file' # The file specified with --defaults-extra-file, if any
    '~/.my.cnf',
    '~/.mylogin.cnf',
]


class MyCnf(object):
    L = logging.getLogger('.'.join([__name__, 'MyCnf']))

    def __init__(self, paths=None):
        if paths is None:
            paths = MYCNF_PATHS

        self.path_stack = list(reversed(paths))
        self.cp = configparser.ConfigParser(allow_no_value=True)

        self.load()

    def valid_path(self, path):
        self.L.debug('checking path {} for validity'.format(path))
        try:
            cnf_stat = os.stat(path)
        except (IOError, OSError):
            self.L.debug('failed to stat path {} (can\'t read/doesn\'t exist?)'.format(path))
            return False

        if cnf_stat.st_mode & stat.S_IWOTH:
            self.L.debug('path {} is world-writable, ignoring'.format(path))
            return False

        return True

    def read(self, path):
        with codecs.open(path) as f:
            for line in f:
                if line.startswith('!'):
                    self.L.debug('found magic directive, line: "{}"'.format(line))

                    self.magic(path, line)
                else:
                    yield line

    def magic(self, sourcefile, line):
        directive, args = line.split(None, 1)
        directive = directive.lstrip('!')
        return {
            'include': self.include,
            'includedir': self.includedir,
        }[directive](sourcefile, args)

    def include(self, source_file, include):
        new_path = os.path.join(os.path.dirname(source_file), include)
        self.L.debug('adding path from source file "{}": {}'.format(
                source_file, new_path))
        self.path_stack.append(new_path)

    def includedir(self, source_file, includedir):
        new_paths = list(glob.iglob(os.path.join(os.path.dirname(source_file), includedir, '*.cnf')))
        self.L.debug('adding paths from source file "{}" found in dir "{}": {}'.format(
                source_file, includedir, new_paths))
        self.path_stack.extend(new_paths)

    def load(self):
        while self.path_stack:
            path = self.path_stack.pop()
            self.L.debug('processing possible path "{}"'.format(path))

            path = os.path.expanduser(path)

            if not self.valid_path(path):
                continue

            self.L.debug('loading/merging file "{}"'.format(path))
            self.read_file(path)

    def read_file(self, path):
        if CP_MODERN:
            self.cp.read_file(self.read(path))
        else:
            buf = io.StringIO('\n'.join(self.read(path)))
            self.cp.readfp(buf)

    def __iter__(self):
        if CP_MODERN:
            return iter(self.cp)
        else:
            return iter(['DEFAULT'] + self.cp.sections())

    def __getitem__(self, key):
        try:
            if CP_MODERN:
                d = dict(self.cp[key])
            else:
                d = dict(self.cp.items(key))
        except KEYERROR_LIKE_OPTIONERRORS as e:
            raise KeyError(str(e))

        return d


if __name__ == '__main__':
    import json
    logging.basicConfig(level=logging.DEBUG)
    mycnf = MyCnf(MYCNF_PATHS)
    print(json.dumps({sec: mycnf[sec] for sec in mycnf}, indent=4))



# coding: utf-8
from __future__ import absolute_import, print_function, unicode_literals

from . import MyCnf, MySqlShim

__all__ = ['MySqlArgs']


class MySqlArgs(object):
    def __init__(self, defaults, mycnfpaths=None):
        mycnf = MyCnf(mycnfpaths)

        for client_key in ['user', 'password', 'host', 'port']:
            try:
                new_value = mycnf['client'][client_key]
            except KeyError:
                continue
            defaults[client_key] = new_value

        self.defaults = defaults

    def inject(self, parser):
        parser.add_argument('-u', '--db-user', type=str,
            default=self.defaults['user'],
            help='Database user (defaulting to "%(default)s")',
        )
        parser.add_argument('-p', '--password', type=str,
            default=self.defaults['password'],
            help='Database password (default empty or as configured with .my.cnf)',
        )
        parser.add_argument('-H', '--host', type=str,
            default=self.defaults['host'],
            help='Database host (defaulting to "%(default)s")',
        )
        parser.add_argument('-P', '--port', type=int,
            default=int(self.defaults['port']),
            help='Database port, ignored for local connections as the UNIX socket '
                 'is used. (defaulting to "%(default)s")',
        )

    def extract(self, args):
        pwd = args.password
        if pwd and len(pwd) > 1 and pwd[0] == pwd[-1] and pwd[0] in '\'\"':
            pwd = pwd[1:-1]

        self.connect_kwargs = {
            'user': args.db_user,
            'passwd': pwd,
            'host': args.host,
            'port': args.port,
        }

    def connect(self):
        return MySqlShim(**self.connect_kwargs)



#!/usr/bin/env python
import sys
import setuptools
from setuptools import setup, find_packages

import hammers

setup(
    name='hammers',
    version=hammers.__version__,
    description='Bag of hammers to fix problems',
    packages=find_packages(),

    author='Nick Timkovich',
    author_email='npt@uchicago.edu',
    url='https://github.org/ChameleonCloud/bag-o-hammers',

    long_description=open('README.rst', 'r').read(),
    keywords=[
        'chameleon-cloud', 'chameleon', 'openstack',
    ],

    entry_points={
        'console_scripts': [
            'neutron-reaper = hammers.scripts.reaper:main',
        ],
    },

    classifiers=[
        'Development Status :: 4 - Beta',
        'Environment :: OpenStack',
        'Intended Audience :: System Administrators',
        'Topic :: Utilities',
    ],

    install_requires=[
        # 'mysqlclient>=1.3.6', # assume this is installed; could also be mysql-python
    ],
)
